from returns.maybe import Maybe, Nothing, Some

from .DomainEventPublisher import DomainEventPublisher

# __all__ = ["DomainService"]


class DomainService:
    """
    Domain services carry domain knowledge that doesnt naturally fit entities and value objects.
    """

    _domain_event_publisher: DomainEventPublisher

    def __init__(
        self,
        domain_event_publisher: Maybe[DomainEventPublisher] = Nothing,
    ) -> None:
        match domain_event_publisher:
            case Some(instance):
                self._domain_event_publisher = instance
            case Maybe.empty:
                self._domain_event_publisher = DomainEventPublisher.instance()
        super().__init__()

    def domain_event_publisher(self) -> DomainEventPublisher:
        return self._domain_event_publisher



from typing import Generic

from returns.maybe import Some
from returns.result import Result
from typing_extensions import Self

from dino_seedwork_be.domain.exceptions import DomainException
from dino_seedwork_be.domain.IdentifiedDomainObject import IdentityType

from .Entity import Entity, OutsideParams

# __all__ = ["AggregateRoot"]


class AggregateRoot(
    Generic[OutsideParams, IdentityType], Entity[OutsideParams, IdentityType]
):
    """Consits of 1+ entities. Spans transaction boundaries."""

    def from_repository(
        self, outside_params: OutsideParams, id: IdentityType
    ) -> Result[Self, DomainException]:
        return self.init(outside_params, Some(id))



from abc import abstractmethod
from typing import Any, Generic, Set, TypeVar

from returns.functions import tap
from returns.future import FutureResult, FutureSuccess
from returns.maybe import Maybe

from dino_seedwork_be.domain.Entity import Entity
from dino_seedwork_be.domain.value_object.AbstractIdentity import \
    AbstractIdentity
from dino_seedwork_be.domain.value_object.NID import NID
from dino_seedwork_be.exceptions import MainException
from dino_seedwork_be.repository.IRepository import Repository
from dino_seedwork_be.utils.functional import maybe_to_future

EntityType = TypeVar("EntityType", bound=Entity)


class MockRepository(Generic[EntityType], Repository[EntityType]):
    _collection: Set[EntityType]

    def __init__(self) -> None:
        self.init_collection()
        super().__init__()

    @abstractmethod
    def init_collection(self):
        pass

    def get_by_id(self, id: NID) -> FutureResult[Maybe[EntityType], Any]:
        found_entity = next(
            (
                dataset
                for dataset in self._collection
                if dataset.identity().value_or(None) == id
            ),
            None,
        )
        # print(
        #     "found entity ",
        #     found_entity,
        #     id,
        #     list(self._collection)[4].identity().unwrap().get_raw(),
        # )
        return FutureSuccess(Maybe.from_optional(found_entity))

    def get_next_id(self, _: bool = False) -> FutureResult[AbstractIdentity, Any]:
        return FutureSuccess(NID(len(self._collection)))

    def add(self, entity: EntityType) -> FutureResult:
        self._collection.add(entity)
        return FutureSuccess(None)

    def save(self, entity: EntityType) -> FutureResult:
        return (
            self.get_by_id(entity.identity().unwrap())
            .bind(maybe_to_future(MainException(code="NOT_FOUND")))
            .map(tap(lambda old_entity: self._collection.remove(old_entity)))
            .map(tap(lambda _: self._collection.add(entity)))
        )

    def remove(self, id: AbstractIdentity) -> FutureResult:
        self._dataset = {
            entity for entity in self._collection if entity.identity() != id
        }
        return FutureSuccess(None)

    def count(self) -> FutureResult[int, Any]:
        return FutureSuccess(len(self._collection))



from typing import (Callable, Optional, ParamSpec, Tuple, Type, TypeVar, Union,
                    overload)

from returns.result import Result, safe

from dino_seedwork_be.domain.exceptions import DomainException

_FuncParams = ParamSpec("_FuncParams")
_ValueType = TypeVar("_ValueType", covariant=True)


@overload
def domain_safe(
    function: Callable[_FuncParams, _ValueType],
) -> Callable[_FuncParams, Result[_ValueType, DomainException]]:
    """Decorator to convert exception-throwing for any kind of Exception."""


@overload
def domain_safe(
    exceptions: Tuple[Type[Exception], ...],
) -> Callable[
    [Callable[_FuncParams, _ValueType]],
    Callable[_FuncParams, Result[_ValueType, DomainException]],
]:
    """Decorator to convert exception-throwing just for a set of Exceptions."""


def domain_safe(  # type: ignore # noqa: WPS234, C901
    function: Optional[Callable[_FuncParams, _ValueType]] = None,
    exceptions: Optional[Tuple[Type[Exception], ...]] = None,
) -> Union[
    Callable[_FuncParams, Result[_ValueType, DomainException]],
    Callable[
        [Callable[_FuncParams, _ValueType]],
        Callable[_FuncParams, Result[_ValueType, DomainException]],
    ],
]:
    def factory(
        inner_function: Callable[_FuncParams, _ValueType],
        inner_exceptions: Tuple[Type[Exception], ...],
    ):
        def decorator(*args: _FuncParams.args, **kwargs: _FuncParams.kwargs):
            return safe(inner_exceptions)(inner_function)(*args, **kwargs).alt(
                lambda e: DomainException(exception=e)
            )

        return decorator

    if callable(function):
        return factory(function, (Exception,))
    if isinstance(function, tuple):
        exceptions = function  # type: ignore
        function = None
    return lambda function: factory(function, exceptions)  # type: ignore



import datetime
from abc import ABC, abstractmethod
from typing import Generic, TypedDict, TypeVar

from returns.maybe import Maybe, Nothing, Some
from returns.result import Result, Success
from typing_extensions import Self

from dino_seedwork_be.fp.domain_safe import domain_safe
from dino_seedwork_be.utils.date import now_utc
from dino_seedwork_be.utils.functional import unwrap

from .IdentifiedDomainObject import IdentifiedDomainObject, IdentityType


class BaseOutsideParams(TypedDict):
    created_at: Maybe[datetime.datetime]
    updated_at: Maybe[datetime.datetime]


OutsideParams = TypeVar("OutsideParams", bound=BaseOutsideParams)


# __all__ = ["Entity", "BaseRawAttributes"]


class Entity(
    ABC, Generic[OutsideParams, IdentityType], IdentifiedDomainObject[IdentityType]
):
    _created_at: Maybe[datetime.datetime] = Nothing
    _updated_at: Maybe[datetime.datetime] = Nothing

    _concurrency_version: int = 0

    def __init__(self, id: Maybe[IdentityType] = Nothing):
        self._concurrency_version = 0
        super().__init__(id)

    def updated_at(self) -> Maybe:
        return self._updated_at

    def created_at(self) -> Maybe:
        return self._created_at

    def set_update_at(self, a_date_time: datetime.datetime) -> Result:
        self._updated_at = Some(a_date_time)
        return Success("OK")

    def set_created_at(self, a_date_time: datetime.datetime) -> Result:
        self._created_at = Some(a_date_time)
        return Success("OK")

    def concurrency_version(self) -> int:
        return self._concurrency_version

    def increase_concurrency_version(self):
        self._concurrency_version += 1

    @classmethod
    @domain_safe
    def init(cls, outside_params: OutsideParams, id: Maybe[IdentityType]) -> Self:
        entity = cls(id)
        entity._created_at = outside_params["created_at"]
        entity._updated_at = outside_params["updated_at"]
        unwrap(entity.from_outside_params(outside_params))
        return entity

    @classmethod
    @domain_safe
    def factory(cls, id: Maybe[IdentityType], *args, **kwargs) -> Self:
        entity = cls(id)
        entity._created_at = Some(now_utc())
        entity._updated_at = Nothing
        entity.create_with_params(*args, **kwargs)
        return entity

    @abstractmethod
    def from_outside_params(self, outside_params: OutsideParams) -> Result:
        pass

    @abstractmethod
    def create_with_params(self, *args, **kwargs) -> Result:
        pass



from typing import Callable, Iterable, List, Tuple, TypeVar

from returns.interfaces.applicative import ApplicativeN
from returns.interfaces.failable import FailableN
from returns.iterables import Fold
from returns.maybe import Maybe, Some
from returns.primitives.hkt import KindN
from returns.result import Result, Success

from dino_seedwork_be.domain.exceptions import DomainException

from .domain_safe import domain_safe
from .list import to_list

InnerValue = TypeVar("InnerValue")
ResultValue = TypeVar("ResultValue")

_FirstType = TypeVar("_FirstType")
_SecondType = TypeVar("_SecondType")
_ThirdType = TypeVar("_ThirdType")
_UpdatedType = TypeVar("_UpdatedType")

_ApplicativeKind = TypeVar("_ApplicativeKind", bound=ApplicativeN)
_FailableKind = TypeVar("_FailableKind", bound=FailableN)


def handle_on_maybe(
    handler: Callable[[InnerValue], Result[None, DomainException]]
) -> Callable[[Maybe[InnerValue]], Result[None, DomainException]]:
    def _handler(inner_value: Maybe[InnerValue]):
        match inner_value:
            case Some(v):
                return handler(v)
            case _:
                return Success(None)

    return _handler


def collect(
    iterable: Iterable[
        KindN[_ApplicativeKind, _FirstType, _SecondType, _ThirdType],
    ],
    acc: KindN[
        _ApplicativeKind,
        "Tuple[_FirstType, ...]",
        _SecondType,
        _ThirdType,
    ],
) -> KindN[_ApplicativeKind, "Tuple[_FirstType, ...]", _SecondType, _ThirdType,]:
    return Fold.collect(iterable, acc)


def collect_all(
    iterable: Iterable[
        KindN[_ApplicativeKind, _FirstType, _SecondType, _ThirdType],
    ],
    acc: KindN[
        _ApplicativeKind,
        "Tuple[_FirstType, ...]",
        _SecondType,
        _ThirdType,
    ],
) -> KindN[_ApplicativeKind, "Tuple[_FirstType, ...]", _SecondType, _ThirdType,]:
    return Fold.collect_all(iterable, acc)


__all__ = ["to_list", "domain_safe", "handle_on_maybe", "collect", "collect_all"]



from collections.abc import Callable
from typing import List, Set, TypeVar

from returns.maybe import Maybe

__all__ = ["remove_none", "unique", "shallow_compare_list"]


def remove_none(aList: List) -> List:
    return list(filter(lambda a: a is not None, aList))


def unique(a_list: List) -> List:
    return list(set(a_list))


def shallow_compare_list(a1: List, a2: List):
    return not any((item not in a2) for item in a1)


T = TypeVar("T")


def get_one(condition: Callable[[T], bool], l: List[T] | Set[T]) -> Maybe[T]:
    return Maybe.from_optional(next((v for v in l if condition(v)), None))
